define([
  "dojo/_base/array",
  "dojo/_base/declare",
  "dojo/_base/event",
  "dojo/_base/lang",
  "dojo/_base/window",
  "dojo/date",
  "dojo/dom-attr",
  "dojo/dom-construct",
  "dojo/dom-style",
  "dojo/on",
  "dojo/text!./templates/DateTimeSelect.html",
  "dijit/_WidgetBase",
  "dijit/_TemplatedMixin",
  "wagby/Select"
], function(array, declare, baseEvent, lang, win, date, domAttr, domConstruct, domStyle, on, template, _WidgetBase, _TemplatedMixin, Select) {
  return declare("wagby.DateTimeSelect", [_WidgetBase, _TemplatedMixin], {
    templateString: template,
    dateFormat: "yyyy-MM-dd",
    yearBegin: null,
    yearRange: 10,
    hourBegin: 0,
    hourRange: 23,
    hourStep: null,
    minuteStep: null,
    secondStep: null,
    rangeType: "",
    onerangeEndName: "",
    termBeginMonth: null,
    subformat: "",
    enterKeyAction: null,
    submitOnChange: false,
    isReadonly: false,

    constructor: function() {
      this._invalidDate = new Date("").toString();
    },

    postCreate: function(){
      this.inherited(arguments);
      if (this.isSelectInputTime()) {
        this.dateFormat = "yyyy-MM-dd HH:mm:ss";
      }
      this.initWidget();
      this._setValueAttr(this.value);
    },

    isSelectInputTime: function() {
      return this.type.indexOf("H") !== -1;
    },

    _setIdAttr: function(id) {
      this.valueNode.setAttribute("id", id);
    },

    _setNameAttr: function(name) {
      if (this.valueNode) {
        this.valueNode.setAttribute("name", name);
      }
    },

    initWidget: function() {
      this.widgets = [];
      if (this.isReadonly) {
        this.createReadOnlyDom();
      } else {
        this.createInputWidgets();
      }
      this.createSubformatDom();
    },

    createReadOnlyDom: function() {
      this.readOnlyDom = domConstruct.create("span", {
        id: "content_" + this.name,
        textContent: this.value,
      }, this.domNode);
    },

    createInputWidgets: function() {
      this.createYearSelect();
      this.createMonthSelect();
      this.createDaySelect();
      this.createHourSelect();
      this.createMinuteSelect();
      this.createSecondSelect();
      this.createOneRangeEndValueNode();
      this.setEvent();
    },

    setEvent: function() {
      if (!this.onchange) {
        return;
      }
      var that = this;
      array.forEach(this.widgets, function(widget) {
        widget.on("change", that.onchange);
      });
    },

    createYearSelect: function() {
      this.yearSelect = this.createSelect({
        refNode: this.yearSelect,
        type: "Y",
        suffix: "_year",
        options: this.createOptions(this.yearBegin, (this.yearBegin + this.yearRange)),
        style: "width: 5em;",
        label: this.termBeginMonth ? res.LABEL_FISCAL_YEAR : res.LABEL_YEAR,
      });
    },

    createMonthSelect: function() {
      this.monthSelect = this.createSelect({
        refNode: this.monthSelect,
        type: "YM",
        suffix: "_month",
        options: this.getMonthOptions(),
        style: "width: 4em;",
        label: res.LABEL_MONTH,
      });
    },

    createDaySelect: function() {
      this.daySelect = this.createSelect({
        refNode: this.daySelect,
        type: "D",
        suffix: "_day",
        options: this.getDayOptions(31),
        style: "width: 4em;",
        label: res.LABEL_DAY,
      });
    },

    createHourSelect: function() {
      this.hourSelect = this.createSelect({
        refNode: this.hourSelect,
        type: "H",
        suffix: "_hour",
        options: this.createOptions(this.hourBegin, this.hourBegin + this.hourRange, this.hourStep),
        style: "width: 4em;",
        label: res.LABEL_HOUR,
      });
    },

    createMinuteSelect: function() {
      this.minuteSelect = this.createSelect({
        refNode: this.minuteSelect,
        type: "HM",
        suffix: "_minute",
        options: this.createOptions(0, 59, this.minuteStep),
        style: "width: 4em;",
        label: res.LABEL_MINUTE,
      });
    },

    createSecondSelect: function() {
      this.secondSelect = this.createSelect({
        refNode: this.secondSelect,
        type: "S",
        suffix: "_second",
        options: this.createOptions(0, 59, this.secondStep),
        style: "width: 4em;",
        label: res.LABEL_SECOND,
      });
    },

    createSelect: function(args) {
      if (this.type.indexOf(args.type) === -1) {
        return;
      }
      var select = new Select({
        id: this.id + args.suffix,
        options: args.options,
        style: args.style,
        onChange: lang.hitch(this, this.setDateFromSelect),
      }).placeAt(args.refNode);
      this.setSelectLabel(select, args.label);
      this.setEnterKeyActionEvent(select);
      this.widgets.push(select);
      return select;
    },

    setEnterKeyActionEvent: function(select) {
      if (!this.enterKeyAction || this.isReadonly) {
        return;
      }
      var keydownFunc = this.getEnterKeyActionFunc();
      if (keydownFunc) {
        select.on("keydown", keydownFunc);
      }
    },

    focusNextElement: function(event) {
      var focusWidget = this.getFocusWidget(event);
      if (focusWidget) {
        focusWidget.focus();
      } else {
        enterKeyHandler.focusNextElement(this, event);
      }
    },

    getFocusWidget: function(event) {
      var id = event.target.id;
      var focusWidget = null;
      for (var i=0,l=this.widgets.length - 1; i<l; i++) {
        var widget = this.widgets[i];
        if (id === widget.id) {
          focusWidget = this.widgets[i + 1];
        }
      }
      return focusWidget;
    },

    getEnterKeyActionFunc: function() {
      var that = this;
      if (this.enterKeyAction === "changetab") {
        return function(event) {
          if (event.keyCode !== 0x0d) return;
          baseEvent.stop(event);
          that.focusNextElement(event);
        }
      } else if (this.enterKeyAction === "ignore") {
        return function(event) {
          if (event.keyCode !== 0x0d) return;
          baseEvent.stop(event);
        }
      } else if (this.enterKeyAction === "default") {
        return function(event) {
          if (event.keyCode !== 0x0d) return;
          baseEvent.stop(event);
          doSubmitForm(that.valueNode.form);
        }
      }
      return null;
    },

    createOneRangeEndValueNode: function() {
      if (!(this.rangeType === "onerange"
            && (this.type === "Y" || this.type === "YM"))) {
        return;
      }
      this.oneRangeEndValueNode = domConstruct.create("input", {
        name: this.onerangeEndName,
        type: "hidden",
      }, this.valueNode, "after");
    },

    setDateTimeSelectValue: function(value) {
      if (!value || value === "") {
        return;
      }
      var dateArray = createDateArray(value, true, this.dateFormat);
      if (!dateArray) {
        return;
      }
      this.setDateSelectValue(dateArray);
      this.setTimeSelectValue(dateArray);
    },

    setDateSelectValue: function(dateArray) {
      var year = dateArray[0];
      var month = dateArray[1];
      var day = dateArray[2];
      var days = this.getDays(year, month);
      this.setDaySelectOptions(days);
      // サーバと送受信する値から年度を考慮して年のリストボックスの値を取得する
      this.setSelectValue(this.yearSelect, this.getFilscalYearSelectValue(year, month));
      this.setSelectValue(this.monthSelect, month);
      this.setSelectValue(this.daySelect, day <= days ? day : days);
      // 単一検索の場合は、隠し項目として保持している終了日の値をセットする
      this.setOneRangeEndValue();
    },

    setTimeSelectValue: function(dateArray) {
      if (!this.isSelectInputTime()) {
        return;
      }
      var hour = dateArray[3] + "";
      var minute = dateArray[4] + "";
      var second = dateArray[5] + "";
      this.setSelectValue(this.hourSelect, hour);
      this.setSelectValue(this.minuteSelect, minute);
      this.setSelectValue(this.secondSelect, second);
    },

    getFilscalYearSelectValue: function(year, month) {
      if (!this.termBeginMonth || this.termBeginMonth === 1 || month >= this.termBeginMonth) {
        return year;
      }
      // サーバと送受信する年は開始月より前の月の場合は翌年となっているので、
      // 年のリストボックスでは前年が選択された状態とする
      return year - 1;
    },

    _isInvalidDate: function(value) {
      return !value || isNaN(value) || typeof value != "object" || value.toString() == this._invalidDate;
    },

    clearDateSelect: function() {
      this.setSelectValue(this.yearSelect, " ");
      this.setSelectValue(this.monthSelect, " ");
      this.setSelectValue(this.daySelect, " ");
      this.setSelectValue(this.hourSelect, " ");
      this.setSelectValue(this.minuteSelect, " ");
      this.setSelectValue(this.secondSelect, " ");
      this.valueNode.value = "";
      this.displaySubformatDate("");
      this.setOneRangeEndValue();
      this.emit("change");
    },

    setOneRangeEndValue: function() {
      if (!this.oneRangeEndValueNode) {
        return;
      }
      if (this.valueNode.value === "") {
        this.oneRangeEndValueNode.value = "";
        return;
      }
      var year = this.getYearValue();
      if (!year) {
        return;
      }
      var month = 12;
      if (this.type === "YM") {
        month = Number(this.monthSelect.get("value"));
      }
      year = this.getFiscalYearValue(year, month);
      var day = this.getDays(year, month);
      this.oneRangeEndValueNode.value = this.getISOFormattedDate(year, month, day);
    },

    setSelectValue: function(dateTimeSelect, value) {
      if (!(dateTimeSelect && dateTimeSelect.set)) {
        return;
      }
      dateTimeSelect.set("_onChangeActive", false);
      dateTimeSelect.set("value", value);
      dateTimeSelect.set("_onChangeActive", true);
    },

    getISOFormattedDateFromSelect: function() {
      var year = this.getYearValue();
      if (!year) {
        return "";
      }
      var month = this.getMonthValue();
      year = this.getFiscalYearValue(year, month);
      var day = this.getDayValue(year, month);
      if (!this.isSelectInputTime()) {
        return this.getISOFormattedDate(year, month, day);
      }
      return this.getISOFormattedDate(year, month, day,
        this.getTimeSelectValue(this.hourSelect),
        this.getTimeSelectValue(this.minuteSelect),
        this.getTimeSelectValue(this.secondSelect));
    },

    getYearValue: function() {
      var yearValue = this.getSelectValue(this.yearSelect);
      if (!yearValue || yearValue === " ") {
        return null;
      }
      return Number(yearValue);
    },

    getMonthValue: function() {
      var monthValue = this.getSelectValue(this.monthSelect);
      if (!monthValue || monthValue === " ") {
        if (this.rangeType === "begin") {
          return this.getTermBeginMonth();
        } else if (this.rangeType === "end") {
          return this.getTermEndMonth();
        } else {
          return 1;
        }
      } else {
        return Number(monthValue);
      }
    },

    getDayValue: function(year, month) {
      var days = this.getDays(year, month);
      var dayValue = this.getSelectValue(this.daySelect);
      if (!dayValue || dayValue === " ") {
        if (this.rangeType === "begin") {
          return 1;
        } else if (this.rangeType === "end") {
          return days;
        } else {
          return 1;
        }
      }
      if (dayValue > days) {
        return days;
      }
      return dayValue;
    },

    getTimeSelectValue: function(timeSelect) {
      var value = this.getSelectValue(timeSelect);
      if (!value || value === " ") {
        return "0";
      }
      return value;
    },

    getISOFormattedDate: function(year, month, day, hour, minute, second) {
      var formattedDate = year + "-" + month + "-" + day;
      if (!this.isSelectInputTime()) {
        return formattedDate
      }
      return formattedDate + " " + (hour || 0) + ":" + (minute || 0) + ":" + (second || 0);
    },

    getSelectValue: function(dateTimeSelect) {
      if (!(dateTimeSelect && dateTimeSelect.get)) {
        return null;
      }
      return dateTimeSelect.get("value");
    },

    getTermBeginMonth: function() {
      if (this.termBeginMonth) {
        return this.termBeginMonth;
      }
      return 1;
    },

    getTermEndMonth: function() {
      var months = 12;
      if (this.termBeginMonth) {
        return (this.termBeginMonth + months - 1) % months;
      }
      return months;
    },

    setDateFromSelect: function() {
      this._setValueAttr(this.getISOFormattedDateFromSelect());
    },

    setSelectLabel: function(dateTimeSelect, label) {
      if (!dateTimeSelect) {
        return;
      }
      domConstruct.create("span", {
        textContent: label,
        'class': "dateTimeSelectLabel",
      }, dateTimeSelect.domNode.parentNode, "after");
    },

    createOptions: function(begin, end, step) {
      step = step || 1;
      var options = [{ label: "", value: " " }];
      for (var i=begin; i<=end; i=i+step) {
        var value = i + "";
        options.push({ label: value, value: value });
      }
      return options;
    },

    getMonthOptions: function() {
      if (!this.termBeginMonth || this.termBeginMonth === 1) {
        return [
          { label: "", value: " " },
          { label: "1", value: "1" },
          { label: "2", value: "2" },
          { label: "3", value: "3" },
          { label: "4", value: "4" },
          { label: "5", value: "5" },
          { label: "6", value: "6" },
          { label: "7", value: "7" },
          { label: "8", value: "8" },
          { label: "9", value: "9" },
          { label: "10", value: "10" },
          { label: "11", value: "11" },
          { label: "12", value: "12" },
        ];
      } else {
        var months = 12;
        var monthOptions = [{ label: "", value: " " }];
        for (var i=this.termBeginMonth,l=this.termBeginMonth + months; i<l; i++) {
          var value = ((i % months) || months);
          monthOptions.push({ label: value + "", value: value });
        }
        return monthOptions;
      }
    },

    getDayOptions: function(days) {
      if (days === 0) {
        return [{ label: "", value: " " }];
      }
      var daysOptions = [
        { label: "", value: " " },
        { label: "1", value: "1" },
        { label: "2", value: "2" },
        { label: "3", value: "3" },
        { label: "4", value: "4" },
        { label: "5", value: "5" },
        { label: "6", value: "6" },
        { label: "7", value: "7" },
        { label: "8", value: "8" },
        { label: "9", value: "9" },
        { label: "10", value: "10" },
        { label: "11", value: "11" },
        { label: "12", value: "12" },
        { label: "13", value: "13" },
        { label: "14", value: "14" },
        { label: "15", value: "15" },
        { label: "16", value: "16" },
        { label: "17", value: "17" },
        { label: "18", value: "18" },
        { label: "19", value: "19" },
        { label: "20", value: "20" },
        { label: "21", value: "21" },
        { label: "22", value: "22" },
        { label: "23", value: "23" },
        { label: "24", value: "24" },
        { label: "25", value: "25" },
        { label: "26", value: "26" },
        { label: "27", value: "27" },
        { label: "28", value: "28" },
      ];
      switch(days) {
        case 29:
          daysOptions.push({ label: "29", value: "29" });
          break;
        case 30:
          daysOptions.push({ label: "30", value: "30" });
          break;
        case 31:
          daysOptions.push({ label: "30", value: "30" });
          daysOptions.push({ label: "31", value: "31" });
          break;
      }
      return daysOptions;
    },

    getTimeOptions: function(range) {
      var options = [{ label: "", value: " " }];
      for (var i=0; i<=range; i++) {
        var value = i + "";
        options.push({ label: value, value: value });
      }
      return options;
    },

    _setValueAttr: function(value) {
      if (!value || value === "" || (value instanceof Date) && this._isInvalidDate(value)) {
        this.clearDateSelect();
        return;
      }
      var dateWithFormat = value;
      try {
        dateWithFormat = _setDateWithFormat(value, this.dateFormat);
      } catch(e) {
      }
      var dateObject;
      try {
        dateObject = createDateObject(value, this.dateFormat);
      } catch(e) {
      }
      if (!dateObject || (dateObject instanceof Date) && this._isInvalidDate(dateObject)) {
        this.clearDateSelect();
      } else {
        this.valueNode.value = dateWithFormat;
        this.displaySubformatDate(dateWithFormat);
        this.setDateTimeSelectValue(dateWithFormat);
      }
      this.emit("change");
    },

    _getValueAttr: function() {
      return this.getISOFormattedDateFromSelect();
    },

    getFilscalYearDate: function(dateObject) {
      var year = this.getFiscalYearValue(dateObject.getFullYear(), dateObject.getMonth() + 1);
      dateObject.setFullYear(year);
      return dateObject;
    },

    setDaySelectOptions: function(days) {
      if (!(this.daySelect && this.daySelect.set)) {
        return;
      }
      this.daySelect.set("options", this.getDayOptions(days));
    },

    getFiscalYearValue: function(year, month) {
      if (!this.termBeginMonth || this.termBeginMonth === 1 || month >= this.termBeginMonth) {
        return year;
      }
      // 開始月より前の月の場合は翌年となる
      return year + 1;
    },

    getDays: function(year, month) {
      var dateObject = createDateObject(this.getFirstDayOfMonth(year, month), this.dateFormat);
      if (!dateObject) {
        return 0;
      }
      return date.getDaysInMonth(dateObject);
    },

    getFirstDayOfMonth: function(year, month) {
      return this.getISOFormattedDate(year, month, 1);
    },

    createSubformatDom: function() {
      if (!this.subformat || this.subformat === "") {
        return;
      }
      this.subformatDom = domConstruct.create("span", {
        id: this.name + "__subformat", style: "white-space: nowrap;"
      }, this.domNode, "after");
      this.displaySubformatDate(this.valueNode.value);
    },

    displaySubformatDate: function(value) {
      if (!this.subformatDom) {
        return;
      }
      var subformatValue = "";
      if (value) {
        var dateObject = createDateObject(value);
        if (dateObject) {
          subformatValue = formatDateObject(dateObject, this.subformat);
        }
      }
      this.subformatDom.textContent = subformatValue;
    },

    _setDisabledAttr: function(value, priorityChange) {
      array.forEach(this.widgets, function(widget) {
        widget.set("disabled", value);
      });
      this.valueNode.disabled = value;
    },

    focus: function() {
      if (this.isReadonly) {
        return;
      }
      this.yearSelect.focusNode.focus();
    },

    reset: function() {
      this.clearDateSelect();
      this.inherited(arguments);
    },

  });
});
